# -*- coding: utf-8 -*-

# state.py
# This file is part of Oqapy serie 3
#
# Author: Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
# Home: http://www.oqapy.eu
# Copyright: 2014-2017 Vincent Vande Vyvre
# Licence: GPL3
#
# Define the status of the application

import logging
lgg = logging.getLogger(__name__)

from PyQt5.QtCore import QObject, Qt, QCoreApplication, pyqtSignal, QSize

class State(QObject):
    _instance = None
    _states = ['empty', 'viewing', 'sorting', 'fullscreen', 'filtering',
               'raw_viewing']
    stateChanged = pyqtSignal(int)
    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(State, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self, core):
        super().__init__()
        self.core = core
        self.ui = core.ui
        self.statusbar = core.ui.statusbar
        self.toolbar = core.ui.toolbar
        self.toolpanel = core.ui.toolpanel
        self.viewers = self.ui.stackedviews
        self.trailers = core.ui.stackedtrailers
        self.current = 0
        self.current_viewer = self.viewers.viewing_area
        self.old_state = 0
        self.dock_visibility = self.toolpanel.isVisible()
        self.trailer_visibility = False
        self.filter_preview = False
        self.is_dual_screen = False

    def set_current_state(self, state, force=False):
        """Sets the current state.

        Args:
        state -- index of the state;
                    0 No viewing, empty list of images
                    1 Mode viewing
                    2 Mode sorting
                    3 Mode full screen
                    4 Mode filtering
                    5 Mode RAW image edition
        force -- if true reset the current state, needed at startup
        """
        lgg.info("Set current state: %s - %s" %(state, self._states[state]))
        self.old_state = self.current
        if state == self.current and not force:
            return

        if self.current == 5 and state != 3:
            if not self.check_raw_status():
                return

            self.close_raw_editing()

        #if state:
            #self.close_current_state(state)

        self.dock_visibility = self.toolpanel.isVisible()
        self.trailer_visibility = self.trailers.isVisible()
        if not state:
            self.set_empty_folder_state(True)

        elif state == 1:
            self.set_viewing_state()

        elif state == 2:
            self.set_sorting_state()

        elif state == 3:
            self.set_full_screen_state()

        elif state == 4:
            self.set_filtering_state(True)

        elif state == 5:
            self.set_raw_viewing_state()

        QCoreApplication.processEvents()
        self.current = state
        self.stateChanged.emit(self.current)

    def reset_old_state(self):
        """Change to the old state.

        """
        lgg.info('Reset old state: %s from state: %s' %(self.old_state, self.current))
        if self.current == 3:
            self.set_transform_state(False)

        elif not self.old_state:
            self.current = 0
            self.set_empty_folder_state()

        elif self.old_state == 1:
            self.current = 1
            self.set_viewing_state()

        elif self.old_state == 2:
            self.current = 1
            self.set_sorting_state()

        self.toolpanel.navigation.set_main_viewer(self.current_viewer)

    def close_current_state(self, new):
        lgg.info('Close state: %s' % self._states[self.current])
        pass

    def get_state_name(self):
        """Returns the name of the current state.

        Returns:
        str(state)
        """
        return self._states[self.current]

    def get_current_viewer(self):
        """Returns the instance of the current active view.

        Returns:
        QGraphicsView instance
        """
        return self.current_viewer

    def set_empty_folder_state(self, empty):
        lgg.info('Source folder is empty: %s' % empty)
        try:
            scene = self.current_viewer.scene()
            scene.clear()
        except AttributeError:
            # At startup the scene doesn't exists
            pass

        if empty:
            self.current_viewer.set_background()
            self.toolpanel.reset_all_toolboxes()
            self.toolpanel.enable_all_toolboxes(False)
            self.trailer_visibility = self.trailers.isVisible()
            self.show_trailer(False)

        self.statusbar.set_current_mode(0)

    def set_viewing_state(self):
        """Change to the viewing state.

        """
        lgg.info('Set viewing state')
        self.viewers.move_viewing_area(0)
        self.viewers.setCurrentIndex(0)
        self.trailers.setCurrentIndex(0)
        self.current_viewer = self.viewers.get_current_viewer()
        self.show_trailer(self.trailer_visibility)
        self.statusbar.set_current_mode(1)
        self.set_target_bar(False)
        #self.toolpanel.reset_all_toolboxes()
        self.toolpanel.enable_all_toolboxes(True)
        self.reset_widgets()
        self.current = 1
        self.core.globalmenu.set_context(self.current)

    def set_sorting_state(self):
        """Change to the sorting mode.

        """
        lgg.info('Set sorting state')
        #self.ui.viewer_1.hide()
        #self.ui.main_tab.show()
        self.viewers.move_viewing_area(1)
        self.viewers.setCurrentIndex(1)
        self.trailers.setCurrentIndex(0)
        self.current_viewer = self.viewers.get_current_viewer()
        self.show_trailer(self.trailer_visibility)
        self.statusbar.set_current_mode(2)
        self.set_target_bar(True)

    def set_full_screen_state(self):
        """Change to the full screen view.

        """
        self.old_state = self.current
        if not self.old_state in (1, 5):
            # Need to change to viewing state before
            self.set_viewing_state()

        lgg.info('Set full screen state, old state: %s' % self.old_state)
        self.set_widgets(False)
        self.current_viewer = self.viewers.get_current_viewer()
        self.current_viewer.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.current_viewer.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.ui.showFullScreen()
        self.current_viewer.setFocus(True)
        self.current = 3

    def exit_fullscreen(self):
        """Back from full screen mode

        """
        lgg.info('Exit full screen state, old state: %s' % self.old_state)
        self.ui.showNormal()
        self.set_widgets(True)
        if self.old_state == 2:
            self.set_sorting_state()
            self.current = 2

        elif self.old_state == 5:
            self.core.set_raw_files_viewing()

        else:
            self.current = 1
        self.core.globalmenu.set_context(self.current)
        self.statusbar.set_current_mode(self.current)

    def set_filtering_state(self, state):
        """Sets or cancel the filtering state.

        Args:
        state -- boolean
        """
        if not state:
            lgg.info('Exit filtering state, old state: %s', self.old_state)
            #self.ui.on_tool_focused(None)
            self.current = self.old_state
            #self.ui.info_bx.setEnabled(True)
            if self.current == 2:
                self.set_sorting_state()

            else:
                self.set_viewing_state()

            #self.reload_current_image()
            self.statusbar.set_current_mode(self.current)
            self.toolbar.widget.global_menu_btn.setEnabled(True)

        else:
            lgg.info('Set filtering state, current state: %s', self.current)
            self.old_state = self.current
            self.trailer_visibility = self.trailers.isVisible()
            self.set_target_bar(False)
            #self.ui.viewer_1.hide()
            #self.ui.main_tab.hide()
            #self.ui.info_bx.setEnabled(False)
            if self.trailer_visibility:
                # Hide cleanely the trailer
                self.ui.toggle_trailer()

            #self.trailers.setCurrentIndex(0)
            #self.ui.edit_area.change_layout_mode(0)
            #self.current_viewer = self.ui.edit_area.viewer
            #self.ui.preview.preview.editing_mode = True
            self.statusbar.set_current_mode(0)
            self.toolbar.widget.global_menu_btn.setEnabled(False)

        self.viewers.viewing_header.set_filtered_mode(state)
        self.enable_navigation_button(not state)
        #self.enable_zoom_buttons(not state)

    def set_raw_viewing_state(self):
        lgg.info('Set raw viewing state, old state: %s', self.old_state)
        #if self.current != 1:
            #self.set_viewing_state()

        self.trailer_visibility = self.trailers.isVisible()
        if not self.current:
            self.set_target_bar()

        self.trailers.setCurrentIndex(1)
        #self.show_trailer(False)
        #self.ui.viewer_1.hide()
        #self.ui.main_tab.hide()
        #self.ui.edit_area.show()
        #self.ui.edit_area.set_raw_mode(True)
        self.show_raw_trailer(True)
        self.enable_tool_boxes(False)
        self.viewers.set_raw_header(True)
        self.toolpanel.set_raw_page_enabled(True)
        self.current = 5
        self.current_viewer = self.viewers.get_current_viewer()
        #self.current_viewer = self.ui.edit_area.viewer
        #self.ui.preview.preview.viewer = self.current_viewer
        #self.ui.preview.preview.editing_mode = True
        #self.statusbar.enable_widgets(False)
        #self.statusbar.set_raw_mode(True)
        #if fls:
            #self.toolpanel.setEnabled(True)
            #self.enable_tool_buttons(True)
            #self.enable_zoom_buttons(True)

    def close_raw_editing(self):
        self.core.rawviewer.close_rawfiles_viewing()
        self.show_raw_trailer(False)
        self.enable_tool_boxes(True)
        self.toolpanel.raw_inquiries.clear()
        self.toolpanel.set_raw_page_enabled(False)
        #self.ui.edit_area.set_raw_mode(False)
        #self.ui.edit_area.hide()
        #self.trailers.setCurrentIndex(1)
        #self.ui.preview.preview.editing_mode = False
        self.core.trailer.is_raw_viewing = False
        self.core.trailer.on_trailer_completed()
        #self.statusbar.set_raw_mode(False)
        #self.statusbar.enable_widgets(True)
        self.viewers.set_raw_header(False)
        self.reset_old_state()
        self.core.viewing.load_picture()

    def enable_tool_boxes(self, enable):
        self.toolpanel.resize_bx.setEnabled(enable)
        self.toolpanel.crop_bx.setEnabled(enable)
        self.toolpanel.straighten_bx.setEnabled(enable)
        self.toolpanel.historic_bx.setEnabled(enable)
        self.toolpanel.contrast_bx.setEnabled(enable)
        self.toolpanel.stretching_bx.setEnabled(enable)
        self.toolpanel.denoising_bx.setEnabled(enable)
        self.toolpanel.grey_map_bx.setEnabled(enable)
        self.toolpanel.toning_bx.setEnabled(enable)
        self.toolpanel.laplace_bx.setEnabled(enable)

    def enable_navigation_button(self, mode):
        """Enables the navigation and rotation buttons.

        Args:
        mode -- boolean, if True the buttons are enabled
        """
        self.toolbar.activate_navigation_widgets(mode)

    def set_main_tab(self, tab):
        """Enables the tool widgets when the main tab has changed.

        Args:
        tab -- index of tab
        """
        self.toolpanel.setEnabled(not tab)
        self.enable_tool_buttons(not tab)
        self.toolbar.magnifier_btn.setEnabled(not tab)
        self.toolbar.zoom_tools.setEnabled(not tab)

        if not tab:
            # Image viewer
            self.show_trailer(self.trailer_visibility)

        else:
            # Target folder
            self.statusbar.enable_widgets(True)
            self.trailer_visibility = self.trailers.isVisible()
            self.show_trailer(False)

    def close_source_folder(self):
        self.set_empty_folder_state(True)

    def show_trailer(self, state=None):
        """Show - hide the trailer.

        Args:
        state -- if None, the old state is used, if True, the trailer is shown
                 otherwise it is hidden
        """
        if self.current == 4:
            self.show_raw_trailer(state)
            return

        lgg.info('show trailer state: %s current: %s' 
                    % (state, self.current))
        if state is None:
            state = not self.trailers.isVisible()
        self.trailers.setVisible(state)
        self.update_trailer_menu(state)

    def show_raw_trailer(self, state=None):
        """Show - hide the trailer.

        Args:
        state -- if None, the old state is used, if True, the trailer is shown
                 otherwise it is hidden
        """
        lgg.info('show raw trailer state: %s current: %s' 
                    % (state, self.current))

        if state is None:
            state = not self.raw_trailer.isVisible()
        self.trailers.setVisible(state)
        self.update_trailer_menu(state)

    def update_trailer_menu(self, state):
        # FIXME replace menu by globalmenu
        return
        self.menus.widget.act_trailer.blockSignals(True)
        self.menus.widget.act_trailer.setChecked(state)
        self.menus.widget.act_trailer.blockSignals(False)
        QCoreApplication.processEvents()
        #self.core.resize()
        #self.core.cfg[1008] = state

    def set_target_bar(self, visible):
        for i in range(self.core.cfg.get(('sorting', 'number'))):
            self.ui.targetarea.targets_buttons[i].setVisible(visible)

    def reset_widgets(self):
        v = self.current_viewer
        self.toolbar.widget.setMaximumSize(QSize(16777215, 52))
        self.trailers.setVisible(self.trailer_visibility)
        self.request_scrollBar(self.core.cfg.get(('style', 'scrlbars')))
        self.toolpanel.navigation.set_main_viewer(v)
        # FIXME
        #self.toolpanel.navigation.preview.editing_mode = False
        try:
            self.toolpanel.navigation.set_image(v, v.current_image)
            self.toolpanel.navigation.update_()
        except AttributeError:
            # No current_image at startup
            pass

    def freeze_tool_boxes(self, caller, freeze=True):
        lgg.info('Freeze tool boxes, caller: %s' %(caller))
        boxes = self.toolpanel.get_tool_boxes()
        if caller is not None:
            caller = caller.parent()

        for box in boxes:
            if box != caller:
                box.setEnabled(not freeze)

    def disable_undo_buttons(self, tool):
        lgg.info('Disable cancel buttons, caller: {0}'.format(tool))
        # Disable all undo buttons
        boxes = self.toolpanel.get_tool_boxes()
        for box in boxes:
            try:
                widget = box.layout.itemAt(1).widget()
            except AttributeError:
                # Case of box is splitter
                continue

            try:
                widget.cancel_btn.setEnabled(False)
            except AttributeError:
                # No cancel button
                continue

        # Enable undo button for the last change
        if tool is not None:
            tool.cancel_btn.setEnabled(True)

    def set_crop_tool_active(self, active):
        self.request_scrollBar(active)
        self.statusbar.set_current_mode(not active)
        self.viewers.viewing_header.enable_layout_buttons(not active)
        if not active:
            self.freeze_tool_boxes(self.toolpanel.crop, False)
            self.toolpanel.navigation.setEnabled(True)

    def set_widgets(self, b):
        """Hide or show widgets when set or unset the full screen mode.

        Args:
        b -- True or False
        """
        self.toolpanel.setVisible(b)
        self.statusbar.widget.setVisible(b)
        self.toolbar.widget.setVisible(b)
        self.viewers.viewing_header.setVisible(b)
        if b:
            b = self.trailer_visibility
        self.trailers.setVisible(b)

    def request_scrollBar(self, b=None):
        """Show - hide scrollBars.

        Args:
        b -- boolean
        """
        if b is None:
            b = self.show_scrollbars

        if b:
            self.current_viewer.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
            self.current_viewer.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)

        else:
            self.current_viewer.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
            self.current_viewer.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        self.show_scrollbars = b

    def reload_current_image(self):
        """Reload the current image after transformations.

        """
        vw = self.core.vwng
        vw.load_current()
        vw.view_current()
        self.core.image_changed(vw.i_pointer, vw.current_qpixmap, 
                                vw.current_zoom, True)

    def clear_tool_boxes(self):
        tbs = [self.ui.image_infos, self.ui.data_view, self.ui.preview,
               self.ui.histogram, self.ui.resizing]

        for tb in tbs:
            tb.clear()

    def check_raw_status(self):
        return True
